@@ -29,6 +29,7 @@ Methods are:
 package XAO::DO::Indexer::Base;
 use strict;
 use Error qw(:try);
+use Encode;
 use XAO::Utils;
 use XAO::Objects;
 use XAO::Projects qw(get_current_project);
@@ -38,10 +39,12 @@ use base XAO::Objects->load(objname => 'Atom');
 
 ### use Data::Dumper;
 
+sub sequential_helper ($$;$$$);
+
 ###############################################################################
 
 use vars qw($VERSION);
-$VERSION='1.01';
+$VERSION=(0+sprintf('%u.%03u',(q$Id: Base.pm,v 1.47 2008/07/06 05:47:54 am Exp $ =~ /\s(\d+)\.(\d+)\s/))) || die "Bad VERSION";
 
 ###############################################################################
 
@@ -65,8 +68,8 @@ sub analyze_text ($$$@) {
         my $kwlist=$self->analyze_text_split($field_num+1,$text);
         my $pos=1;
         foreach my $kw (@$kwlist) {
-            $kw_data->{count_uid}->{$unique_id}->{$kw}++;
-            if(! exists $kw_data->{ignore}->{$kw}) {
+            $kw_data->{'count_uid'}->{$unique_id}->{$kw}++;
+            if(! exists $kw_data->{'ignore'}->{$kw}) {
                 my $kwt=$kw_info->{lc($kw)}->{$unique_id};
                 if(! $kwt->[$field_num]) {
                     $kw_info->{lc($kw)}->{$unique_id}->[$field_num]=[ $pos ];
@@ -95,20 +98,33 @@ sub analyze_text_split ($$$) {
 
 ###############################################################################
 
-sub ignore_limit ($) {
+sub config_param ($$) {
+    my ($self,$param,$default)=@_;
+
     my $config=get_current_project();
-    my $limit;
+    my $value;
     if($config) {
-        my $self=shift;
-        my $args=get_args(\@_);
-        if($args->{index_object}) {
-            my $index_id=$args->{index_object}->container_key;
-            $limit=$config->get("/indexer/$index_id/ignore_limit");
-        }
-        $limit||=$config->get('/indexer/default/ignore_limit');
+        my $index_id=$self->{'index_id'} ||
+            throw $self "config_param - no index_id (improper creation?)";
+        $value=$config->get("/indexer/$index_id/$param");
+        $value||=$config->get("/indexer/default/$param");
     }
 
-    return $limit || 500;
+    return $value || $default;
+}
+
+###############################################################################
+
+sub ignore_limit ($) {
+    my $self=shift;
+    return $self->config_param('ignore_limit',500);
+}
+
+###############################################################################
+
+sub commit_interval ($) {
+    my $self=shift;
+    return $self->config_param('commit_interval',0);
 }
 
 ###############################################################################
@@ -117,7 +133,7 @@ sub init ($%) {
     my $self=shift;
     my $args=get_args(\@_);
 
-    my $index_object=$args->{index_object} ||
+    my $index_object=$args->{'index_object'} ||
         throw $self "init - no 'index_object'";
 
     dprint ref($self)."::init - XXX, nothing's in here yet....";
@@ -129,13 +145,13 @@ sub search ($%) {
     my $self=shift;
     my $args=get_args(\@_);
 
-    my $index_object=$args->{index_object} ||
+    my $index_object=$args->{'index_object'} ||
         throw $self "search - no 'index_object'";
 
     eval 'use Compress::LZO';
     dprint "No Compress::LZO, comression won't work" if $@;
 
-    my $str=$args->{search_string};
+    my $str=$args->{'search_string'};
     $str='' unless defined $str;
     $str=~s/^\s*(.*?)\s*$/$1/sg;
     if(!length($str)) {
@@ -143,18 +159,29 @@ sub search ($%) {
         return [ ];
     }
 
-    my $ordering=$args->{ordering} ||
+    my $ordering=$args->{'ordering'} ||
         throw $self "search - no 'ordering'";
-    my $ordering_seq=$self->get_orderings->{$ordering}->{seq} ||
+    my $ordering_seq=$self->get_orderings->{$ordering}->{'seq'} ||
         throw $self "search - no sequence in '$ordering' ordering";
 
     ##
     # Optional hash reference to be filled with statistics
     #
-    my $rcdata=$args->{rcdata};
-    $rcdata->{ignored_words}={ } if $rcdata;
+    my $rcdata=$args->{'rcdata'};
+    $rcdata->{'ignored_words'}={ } if $rcdata;
 
-    dprint "Searching for '$str' (ordering=$ordering, seq=$ordering_seq)";
+    ### dprint "Searching for '$str' (ordering=$ordering, seq=$ordering_seq)";
+
+    ##
+    # Preparing spellchecker if needed
+    #
+    my $use_spellchecker=$rcdata && $self->config_param('use_spellchecker');
+    my $spellchecker;
+    if($use_spellchecker) {
+        $spellchecker=$self->get_spellchecker;
+        $spellchecker->switch_index($index_object->container_key);
+        $rcdata->{'spellchecker_words'}={ };
+    }
 
     ##
     # We cache ignored words. Cache returns a hash reference with
@@ -163,12 +190,11 @@ sub search ($%) {
     my $i_cache=get_current_project()->cache(
         name        => 'indexer_ignored',
         coords      => [ 'index_id' ],
-        expire      => 3,
+        expire      => 60,
         retrieve    => sub {
             my $index_list=shift;
             my $args=get_args(\@_);
-            my $index_id=$args->{index_id};
-            dprint "CACHE: retrieving ignored words for index $index_id";
+            my $index_id=$args->{'index_id'};
             my $ign_list=$index_list->get($index_id)->get('Ignore');
             my %ignored=map {
                 $ign_list->get($_)->get('keyword','count');
@@ -197,10 +223,14 @@ sub search ($%) {
             push(@simple,$s->[0]);
         }
         else {
+            if($spellchecker) {
+                my $pairs=$spellchecker->suggest_replacements(join(' ',@$s));
+                @{$rcdata->{'spellchecker_words'}}{keys %$pairs}=values %$pairs;
+            }
             my @t=map {
                 if(exists $ignored->{$_}) {
                     if($rcdata) {
-                        $rcdata->{ignored_words}->{$_}=$ignored->{$_};
+                        $rcdata->{'ignored_words'}->{$_}=$ignored->{$_};
                     }
                     undef;
                 }
@@ -223,10 +253,14 @@ sub search ($%) {
     ##
     # Simple words
     #
+    if($spellchecker) {
+        my $pairs=$spellchecker->suggest_replacements($str);
+        @{$rcdata->{'spellchecker_words'}}{keys %$pairs}=values %$pairs;
+    }
     push(@simple,map {
         if(exists $ignored->{$_}) {
             if($rcdata) {
-                $rcdata->{ignored_words}->{$_}=$ignored->{$_};
+                $rcdata->{'ignored_words'}->{$_}=$ignored->{$_};
             }
             ();
         }
@@ -235,6 +269,14 @@ sub search ($%) {
         }
     } @{$self->analyze_text_split(0,$str)});
 
+    ##
+    # If we are asked to provide data, storing splitted words.
+    #
+    if($rcdata) {
+        $rcdata->{'words_single'}=\@simple;
+        $rcdata->{'words_multi'}=\@multi;
+        $rcdata->{'results_count'}=0;
+    }
     ### dprint Dumper(\@multi),Dumper(\@simple);
 
     ##
@@ -267,7 +309,14 @@ sub search ($%) {
     ##
     # Joining all results together
     #
-    return XAO::IndexerSupport::sorted_intersection(@results);
+    if($rcdata) {
+        my $sr=XAO::IndexerSupport::sorted_intersection(@results);
+        $rcdata->{'results_count'}=scalar(@$sr);
+        return $sr;
+    }
+    else {
+        return XAO::IndexerSupport::sorted_intersection(@results);
+    }
 }
 
 ###############################################################################
@@ -284,11 +333,21 @@ sub search_multi ($$$$) {
         last unless defined $kw;
         my $sr=$data_list->search('keyword','eq',$kw);
         return [ ] unless @$sr;
+
+        use bytes;
+
         my $r=$data_list->get($sr->[0])->get("idpos_$oseq");
-        if(unpack('w',$r) == 0) {
-            my $zz=substr($r,1);
-            $r=Compress::LZO::decompress($zz);
+
+        my $compression_flag=unpack('w',$r);
+        if(defined $compression_flag && $compression_flag == 0) {
+
+            # The temporary variable is required. See the note on the
+            # other decompress() use.
+            #
+            my $c=substr($r,1);
+            $r=Compress::LZO::decompress($c);
         }
+
         $rawdata{$kw}=$r;
     }
 
@@ -305,24 +364,30 @@ sub search_simple ($$$$) {
 
     my $iddata=$data_list->get($sr->[0])->get("id_$oseq");
 
-    ##
-    # Decompressing if required
-    #
-    if(unpack('w',$iddata) == 0) {
-        my $zz=substr($iddata,1);
-        $iddata=Compress::LZO::decompress($zz);
-        if(!defined $iddata) {
-            eprint "Can't decompress data for kw='$keyword', sorting='$oseq'";
-            return [ ];
-        }
-    }
-
     ##
     # Sometimes the data gets damaged or is in the process of being
     # updated. This can lead to unparsable results.
     #
     my $result;
     try {
+        use bytes;
+
+        my $compression_flag=unpack('w',$iddata);
+        if(defined $compression_flag && $compression_flag==0) {
+
+            # Directly passing substr(..) into decompress fails in perl
+            # 5.22 with "buffer parameter is not a SCALAR". Likely this
+            # is because substr is actually an lvalue and not a true
+            # scalar. Worked fine in earlier versions of perl. Putting
+            # it in a temporary variable fixes that.
+            #
+            my $c=scalar(substr($iddata,1));
+            $iddata=Compress::LZO::decompress($c);
+
+            defined $iddata ||
+                throw $self "Can't decompress data";
+        }
+
         $result=[ unpack('w*',$iddata) ];
     }
     otherwise {
@@ -330,16 +395,221 @@ sub search_simple ($$$$) {
         eprint "Bad indexer data for kw='$keyword', sorting='$oseq': $e";
         $result=[ ];
     };
+
     return $result;
 }
 
 ###############################################################################
 
+sub suggest_alternative ($%) {
+    my $self=shift;
+    my $args=get_args(\@_);
+
+    my $index_object=$args->{'index_object'} ||
+        throw $self "search - no 'index_object'";
+
+    my $rcdata=$args->{'rcdata'} ||
+        throw $self "suggest_alternatives - need rcdata";
+
+    my $query=$args->{'search_string'} ||
+        throw $self "suggest_alternatives - need search_string";
+
+    my $spwords=$rcdata->{'spellchecker_words'};
+    return '' unless $spwords;
+
+    ##
+    # This can be used to improve results when using a generic
+    # dictionary on a small dataset, not a dictionary based on the
+    # actual index content (which guarantees that we already get only
+    # valid substitutes).
+    #
+    my %wcounts;
+    my $max_alt_words=$self->config_param('spellchecker/max_alt_words') || 0;
+    if($max_alt_words) {
+        dprint "Using 'max_alt_words' slows down suggestion, consider building a custom dictionary";
+        my $data_list=$index_object->get('Data');
+        foreach my $word (keys %$spwords) {
+            my $alist=$spwords->{$word};
+            my @newlist;
+            for(my $i=0; $i<$max_alt_words && $i<@$alist; ++$i) {
+                my $altword=lc($alist->[$i]);
+                ### dprint "Trying word '$word' -> '$altword'";
+
+                my @aw=$self->analyze_text_split(0,$altword);
+                my $found=0;
+                foreach my $aw (@aw) {
+                    my $sr=$data_list->search('keyword','eq',$aw);
+                    next unless @$sr;
+                    ++$found;
+                    $wcounts{$aw}=scalar(@$sr);
+                }
+
+                if($found && $found==scalar(@aw)) {
+                    push(@newlist,$altword);
+                }
+            }
+            $spwords->{$word}=\@newlist;
+        }
+    }
+
+    ##
+    # Ordering the list of substitutions to try on the original query.
+    # Fuzzy area, there could be multiple algorithms to do that.
+    #
+    my $algorithm=$self->config_param('spellchecker/algorithm') || 'sequential';
+    my $max_alt_searches=$self->config_param('spellchecker/max_alt_searches') || 15;
+    my $max_alt_results=$self->config_param('spellchecker/max_alt_results') || 2;
+    my $max_result_distance=$self->config_param('spellchecker/max_result_distance') || 5;
+    ### dprint ".algorithm=$algorithm max_alt_results=$max_alt_results max_alt_searches=$max_alt_searches";
+
+    ##
+    # In 'sequential' we scan through all words from most likely to
+    # least likely, with all possible variations. If we have these variants:
+    # A->(B,C)  K->(L,M)  X->(Y)
+    # then the following will be tried
+    # (A-B) (K-L) (X-Y) (A-B,K-L) (A-B,X-Y) (K-L,X-Y) (A-B,K-L,X-Y) (A-B,K-M)
+    # (A-C,K-L) (A-C,X-Y) (K-M,X-Y) (A-C,K-M) (A-C,K-L,X-Y), etc..
+    #
+    # The idea is that we assign weight proportional to the distance
+    # from the top of the list, make lists and then sort on the weight.
+    #
+    my @jobs;
+    if($algorithm eq 'sequential') {
+
+        if($self->config_param('spellchecker/try_word_removals')) {
+            foreach my $word (keys %$spwords) {
+                my $n=scalar(@{$spwords->{$word}});
+                if($n>2) {
+                    splice(@{$spwords->{$word}},2,0,'');
+                }
+                elsif($n) {
+                    push(@{$spwords->{$word}},'');
+                }
+            }
+        }
+
+        my @list;
+        sequential_helper($spwords,\@list);
+
+        ### use Data::Dumper;
+        ### dprint Dumper(\@list);
+
+        foreach my $elt (@list) {
+            my $distance=0;
+            my $count=0;
+            foreach my $pair (@{$elt->{'job'}}) {
+                $distance+=$pair->[2];
+                ++$count;
+            }
+            $elt->{'distance'}=$distance+$count-1;
+        }
+
+        @jobs=map {
+            {
+                pairs       => [ map { ($_->[0],$_->[1]) } @{$_->{'job'}} ],
+                distance    => $_->{'distance'},
+            }
+        } sort {
+            $a->{'distance'} <=> $b->{'distance'} ||
+            scalar(@{$b->{'job'}}) <=> scalar(@{$a->{'job'}})
+        } @list;
+
+        ### use Data::Dumper;
+        ### dprint Dumper(\@jobs);
+
+        if(@jobs>$max_alt_searches) {
+            splice(@jobs,$max_alt_searches);
+        }
+    }
+
+    ##
+    # In 'bycount' we keep completely dropping least likely words. Does
+    # not work very well, probably should be removed altogether as unusable.
+    #
+    ### elsif($algorithm eq 'bycount') {
+    ###     for(my $i=0; %words && $i<$max_alt_searches; ++$i) {
+    ###         my @wlist=sort { $words{$b}->[0]->[1] <=> $words{$a}->[0]->[1] } keys %words;
+    ###         my @pairs;
+    ###         my $have_difference;
+    ###         foreach my $word (@wlist) {
+    ###             my $altword=$words{$word}->[0]->[0];
+    ###             $have_difference=1 if $altword ne $word;
+    ###             push(@pairs,$word => $altword);
+    ###         }
+    ###         push(@jobs,{
+    ###             pairs       => @pairs,
+    ###             distance    => 1,
+    ###         }) if $have_difference;
+    ###         my $word=$wlist[$#wlist];
+    ###         if(@{$words{$word}}>1) {
+    ###             shift(@{$words{$word}});
+    ###         }
+    ###         else {
+    ###             delete $words{$word};
+    ###         }
+    ###     }
+    ### }
+    else {
+        throw $self "suggest_alternative - '$algorithm' algorithm is not supported";
+    }
+
+    # Now building other alternative strings and returning them in order.
+    #
+    my $results_count=$rcdata->{'results_count'} || 0;
+    my @alts;
+    for(my $i=0; $i<@jobs; ++$i) {
+        my $distance=$jobs[$i]->{'distance'} || 1;
+        $distance<=$max_result_distance || next;
+
+        my $newq=$query;
+        my $pairs=$jobs[$i]->{'pairs'};
+        my @finalpairs;
+        for(my $j=0; $j<@$pairs; $j+=2) {
+            my $word=$pairs->[$j];
+            my $altword=$pairs->[$j+1];
+            if($word ne $altword) {
+                $newq=~s/\b$word\b/$altword/ig;
+                $newq=~s/^\s*(.*?)\s*$/$1/s;
+                $newq=~s/\s{2,}/ /s;
+                push(@finalpairs,[ $word, $altword ]);
+            }
+        }
+        next if $newq =~ /^\s+$/s;
+
+        dprint "Trying query '$newq' instead of '$query'";
+        my $sr=$self->search(
+            index_object    => $index_object,
+            search_string   => $newq,
+            ordering        => $args->{'ordering'},
+        );
+        my $newcount=@$sr;
+        if($newcount && $newcount>=$results_count) {
+            dprint "Got a match on '$newq' ($newcount)";
+            push(@alts,{
+                query       => $newq,
+                pairs       => \@finalpairs,
+                results     => $args->{'need_results'} ? $sr : undef,
+                count       => $newcount,
+                distance    => $distance,
+            });
+
+            last if scalar(@alts)>=$max_alt_results;
+        }
+    }
+
+    # Storing all variants and returning the first one.
+    #
+    $rcdata->{'spellchecker_alternatives'}=\@alts;
+    return @alts ? $alts[0]->{'query'} : undef;
+}
+
+###############################################################################
+
 sub update ($%) {
     my $self=shift;
     my $args=get_args(\@_);
 
-    my $index_object=$args->{index_object} ||
+    my $index_object=$args->{'index_object'} ||
         throw $self "update - no 'index_object'";
 
     ##
@@ -375,20 +645,30 @@ sub update ($%) {
     # Maximum number of matches for a word for it to be completely
     # ignored.
     #
-    my $ignore_limit=$self->ignore_limit($args);
+    my $ignore_limit=$self->ignore_limit;
+
+    ##
+    # Maximum number of keywords updated before a forced transaction
+    # commit. This helps to deal with otherwise huge transaction that
+    # have to be cached before being moved into binlogs for slave
+    # updates in MySQL -- it is my understanding that it will require at
+    # least tripple write, and at least one atomic huge write. Better
+    # not do that.
+    #
+    my $commit_interval=$self->commit_interval;
 
     ##
     # If that's a partial update and we don't have any IDs to update
     # we return immediately. Otherwise proceed even with empty set to
     # remove records from the index.
     #
-    my $is_partial=$cinfo->{partial};
-    my $coll_ids=$cinfo->{ids};
+    my $is_partial=$cinfo->{'partial'};
+    my $coll_ids=$cinfo->{'ids'};
     my $coll_ids_total=scalar(@$coll_ids);
     if($is_partial && !$coll_ids_total) {
         return 0;
     }
-    my $coll=$cinfo->{collection};
+    my $coll=$cinfo->{'collection'};
 
     ##
     # For partial indexes we pre-load ignored words. Otherwise, as the
@@ -444,6 +724,8 @@ sub update ($%) {
     my $now=time;
 
     $index_object->glue->transact_begin;
+    my $transact_count=0;
+
     my $data_list=$index_object->get('Data');
     my $max_kw_length=$data_list->get_new->describe('keyword')->{'maxlength'};
     my $ni=$ignore_list->get_new;
@@ -452,6 +734,12 @@ sub update ($%) {
     my %o_prepare;
     my %o_finish;
 
+    ##
+    # Data length limits
+    #
+    my $maxlen_id=$data_list->get_new->describe('id_1')->{'maxlength'};
+    my $maxlen_idpos=$data_list->get_new->describe('idpos_1')->{'maxlength'};
+
     ##
     # As the list can potentially be huge, we go one ordering at a time,
     # thus probably re-doing the same record -- which is fine for large
@@ -509,7 +797,7 @@ sub update ($%) {
             # ignore it and issue a warning. Should not happen often --
             # if it does the field length should be extended.
             #
-            if(length($kw) > $max_kw_length) {
+            if(length(Encode::encode('utf8',$kw)) > $max_kw_length) {
                 eprint "Keyword '$kw' is longer than $max_kw_length, ignored";
                 next;
             }
@@ -523,10 +811,24 @@ sub update ($%) {
                 dprint "..$count/$kw_total (".localtime($tstamp)." - ".localtime($tstamp_end).")";
             }
 
+            ##
+            # If it's time to commit - committing and starting a new
+            # transaction. Yes, if we fail we can end up with a
+            # partially update data, but nothing terribly bad -- it'll
+            # join fine on next update anyway and will be usable in
+            # between.
+            #
+            if($commit_interval && ++$transact_count>$commit_interval) {
+                dprint "...forced transaction commit";
+                $index_object->glue->transact_commit;
+                $index_object->glue->transact_begin;
+                $transact_count=0;
+            }
+
             ##
             # Using md5 based IDs
             #
-            my $kwmd5=md5_base64($kw);
+            my $kwmd5=md5_base64(Encode::encode_utf8($kw));
             $kwmd5=~s/\W/_/g;
             my $data_obj;
             if($data_list->exists($kwmd5)) {
@@ -649,6 +951,14 @@ sub update ($%) {
                 #dprint "COMPR.aft: id=".length($iddata)." idpos=".length($posdata);
             }
 
+            ##
+            # Checking length before storing
+            #
+            length($iddata) <= $maxlen_id ||
+                throw $self "update - id data too long (".length($iddata).">$maxlen_id), kw='$kw', count=".$kw_data{'counts'}->{$kw};
+            length($posdata) <= $maxlen_idpos ||
+                throw $self "update - pos data too long (".length($posdata).">$maxlen_idpos), kw='$kw', count=".$kw_data{'counts'}->{$kw};
+
             ##
             # Storing
             #
@@ -729,17 +1039,116 @@ sub get_collection ($%) {
 sub finish_collection ($%) {
     my $self=shift;
     my $args=get_args(\@_);
-    my $cinfo=$args->{collection_info} ||
+    my $cinfo=$args->{'collection_info'} ||
         throw $self "finish_collection - no 'collection_info' given";
 
     ##
     # If it's partial and we're called -- we throw an exception to
     # indicate that this method must be overriden.
     #
-    if($cinfo->{partial}) {
+    if($cinfo->{'partial'}) {
         throw $self "finish_collection - implementation required for partial collections";
     }
 }
 
+###############################################################################
+
+sub build_dictionary ($%) {
+    my $self=shift;
+    my $args=get_args(\@_);
+
+    my $index_object=$args->{'index_object'} ||
+        throw $self "build_dictionary - no 'index_object'";
+
+    my $spellchecker=$self->get_spellchecker;
+
+    $spellchecker->switch_index($self->{'index_id'});
+
+    my $wlist=$spellchecker->dictionary_create;
+    return unless $wlist;
+
+    my $dict_req_count=$self->config_param('spellchecker/dictionary_req_count') || 3;
+    my $dict_req_length=$self->config_param('spellchecker/dictionary_req_length') || 1;
+
+    my $data_list=$index_object->get('Data');
+    my @data_keys=$data_list->keys;
+    my $datacount=0;
+    my $datatotal=scalar(@data_keys);
+    my $wcount=0;
+    dprint ".words from search index ($datatotal)";
+    foreach my $data_id (@data_keys) {
+        dprint "..$datacount/$datatotal, word count $wcount" if (++$datacount%5000)==0;
+
+        my ($keyword,$count)=$data_list->get($data_id)->get('keyword','count');
+
+        next unless $count>=$dict_req_count &&
+                    length($keyword)>=$dict_req_length;
+
+        $wcount=$spellchecker->dictionary_add($wlist,$keyword,$count);
+    }
+
+    $data_list=$index_object->get('Ignore');
+    @data_keys=$data_list->keys;
+    $datacount=0;
+    $datatotal=scalar(@data_keys);
+    dprint ".words from ignore list ($datatotal)";
+    foreach my $data_id (@data_keys) {
+        dprint "..$datacount/$datatotal, word count $wcount" if (++$datacount%5000)==0;
+
+        my ($keyword,$count)=$data_list->get($data_id)->get('keyword','count');
+
+        next unless $count>=$dict_req_count &&
+                    length($keyword)>=$dict_req_length;
+
+        $wcount=$spellchecker->dictionary_add($wlist,$keyword,$count);
+    }
+
+    $spellchecker->dictionary_close($wlist);
+}
+
+###############################################################################
+
+sub sequential_helper ($$;$$$) {
+    my ($words,$list,$base,$level,$ixhash)=@_;
+
+    $base||=[ ];
+    $level||=0;
+    $ixhash||={ };
+
+    for(my $i=0; $i<5; ++$i) {
+        my $newlevel=$level+$i;
+        foreach my $word (keys %$words) {
+            next if grep { $_->[0] eq $word } @$base;
+            my $altlist=$words->{$word};
+            next if $newlevel>=@$altlist;
+            my $altword=$altlist->[$newlevel];
+            next if $altword eq $word;
+
+            my @newbase=(@$base, [ $word, $altword, $newlevel ]);
+
+            my $ix=join('|',sort map { $_->[0].'-'.$_->[1] } @newbase);
+            next if exists $ixhash->{$ix};
+            $ixhash->{$ix}=1;
+
+            push(@$list,{
+                job         => \@newbase,
+            });
+
+            scalar(@$list)<100 &&
+                sequential_helper($words,$list,\@newbase,$newlevel,$ixhash);
+        }
+    }
+}
+
+###############################################################################
+
+sub get_spellchecker ($) {
+    my $self=shift;
+
+    return XAO::Objects->new(
+        objname     => $self->config_param('spellchecker/objname') || 'SpellChecker::Aspell',
+    );
+}
+
 ###############################################################################
 1;
